local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 9
local LibUnitID,oldminor = LibStub:NewLibrary(MAJOR,MINOR)
if (not LibUnitID) then return end -- already loaded, no upgrade necessary

-- localize upvalues
local error,next,pairs,pcall,setmetatable,type,wipe = error,next,pairs,pcall,setmetatable,type,wipe
local API_UnitGUID = UnitGUID

-- dump data from old version, if any
local olddata = {}
if (type(LibUnitID.__dump_for_upgrade) == "function") then
	olddata.MINOR = oldminor
	if (oldminor < 9) then
		local frame,listeners,tests,updates = LibUnitID:__dump_for_upgrade()
		olddata = { frame=frame, tagListeners=listeners, logTests=tests, logUpdates=updates }
	else
		olddata = LibUnitID:__dump_for_upgrade()
	end
end

-- hash table: tags that can be tracked with events
local tagSupported = {
	--player not needed since it never changes
	vehicle=1,vehicletarget=1,
	pet=1,pettarget=1,
	target=1,targettarget=1,
	focus=1,focustarget=1
}
for i=1,5 do tagSupported["party"..i]=1; tagSupported["partypet"..i]=1; tagSupported["partypet"..i.."target"]=1; end
for i=1,40 do tagSupported["raid"..i]=1; tagSupported["raidpet"..i]=1; tagSupported["raidpet"..i.."target"]=1; end

-- hash table: tags that are the pet of (thus dependent on) other tags
local tagPet = {
	player="pet"
	--player="vehicle" is equally valid but doesn't fit the scheme very well, so there's a special check for that in UNIT_PET
	--target and focus get UNIT_PET events, but targetpet and focuspet aren't valid unitIDs.  boggle
}
for i=1,5 do tagPet["party"..i]="partypet"..i; end
for i=1,40 do tagPet["raid"..i]="raidpet"..i; end

-- hash table: tags that are the target of (thus dependent on) other tags
local tagTarget = {
	--player="target" not needed since we use PLAYER_TARGET_CHANGED instead of UNIT_TARGET(player)
	vehicle="vehicletarget",
	pet="pettarget",
	target="targettarget",
	focus="focustarget"
}
for i=1,5 do tagTarget["party"..i]="party"..i.."target"; tagTarget["partypet"..i]="partypet"..i.."target"; end
for i=1,40 do tagTarget["raid"..i]="raid"..i.."target"; tagTarget["raidpet"..i]="raidpet"..i.."target"; end

-- hash table: tags that depend on another tag
local tagParent = {
	--target="player" not needed since player never changes
	--vehicle="player", ditto
	vehicletarget="vehicle",
	pettarget="pet",
	targettarget="target",
	focustarget="focus"
}
for i=1,5 do tagParent["party"..i.."target"]="party"..i; tagParent["partypet"..i]="party"..i; tagParent["partypet"..i.."target"]="partypet"..i; end
for i=1,40 do tagParent["raid"..i.."target"]="raid"..i; tagParent["raidpet"..i]="raid"..i; tagParent["raidpet"..i.."target"]="raidpet"..i; end

-- hash table: tags that depend on the party composition, for batch registration
local partyTags = {}
for i=1,5 do partyTags["party"..i]=1; partyTags["party"..i.."target"]=1; partyTags["partypet"..i]=1; partyTags["partypet"..i.."target"]=1; end

-- hash table: tags that depend on the raid composition, for batch registration
local raidTags = {}
for i=1,40 do raidTags["raid"..i]=1; raidTags["raid"..i.."target"]=1; raidTags["raidpet"..i]=1; raidTags["raidpet"..i.."target"]=1; end

-- runtime data
local frame = olddata.frame or CreateFrame("Frame", nil, UIParent)
local tagTracked = {} -- tagTracked[unitid] = true/false
local tagID = {} -- tagID[unitid] = guid
local tagFunctions = {} -- tagFunctions[unitid][cb] = true
local tagMethods = {} -- tagMethods[unitid][handler][cb] = true
local logTests = olddata.logTests or 0
local logUpdates = olddata.logUpdates or 0
local logQueries = olddata.logQueries or 0
local queueFunctions = {} -- queueFunctions[cb] = true
local queueMethods = {} -- queueMethods[handler][cb] = true

-- the delta table
--- The "delta" table is provided to all callbacks on any UnitID change.
-- @name delta
-- @class table
-- This table lists the UnitIDs that changed during a single game event,
-- and the new GUIDs associated with each.  For efficiency, the table is
-- shared and provided to every callback involved during a game event; for
-- safety, that table may not be modified (it '''__index'''es the real table,
-- and errors on '''__newindex''').  This means you cannot process it using
-- '''pairs()''', '''ipairs()''' or '''next()'''.  Instead, index '''0''' of
-- the table will contain the number of affected UnitIDs (the equivalent of
-- the '''#''' operator), and indecies '''1'''-'''#''' will contain those
-- UnitIDs.  In addition, each affected UnitID is a string key of the table,
-- whose value is that UnitID's new GUID (which may be nil, if the unit no
-- longer exists).  Please see the project documentation for code examples to
-- process this table in your callback functions.
-- @field 0 The number of UnitIDs that changed.
-- @field (n) (Integer) For n = 1 .. delta[0], delta[n] is a UnitID that changed.
-- @field (unitid) (String) For unitid in { delta[1] .. delta[delta[0]] }, delta[unitid] is the new GUID associated with the UnitID.
local delta = {}
local deltaSize = 0
local deltaWrapper = setmetatable({}, {
	__index = delta,
	__newindex = function() error("cannot modify shared table") end,
	__metatable = false}
)


--[[ INTERNAL FUNCTIONS ]]--

local function update_events()
	local tagTracked,pairs,frame = tagTracked,pairs,frame
	-- identify necessary events
	local p_t_c = (tagTracked.target)
	local p_f_c = (tagTracked.focus)
	local u_p = (tagTracked.vehicle or tagTracked.vehicletarget or tagTracked.pet)
	if (not u_p) then
		for parent,tag in pairs(tagPet) do
			if (tagTracked[tag]) then
				u_p = true
				break
			end
		end
	end
	local u_t = (tagTracked.targettarget or tagTracked.focustarget or tagTracked.pettarget)
	if (not u_t) then
		for parent,tag in pairs(tagTarget) do
			if (tagTracked[tag]) then
				u_t = true
				break
			end
		end
	end
	local p_m_c = false
	for tag,_ in pairs(partyTags) do
		if (tagTracked[tag]) then
			p_m_c = true
			break
		end
	end
	local r_r_u = false
	for tag,_ in pairs(raidTags) do
		if (tagTracked[tag]) then
			r_r_u = true
			break
		end
	end
	-- register necessary events
	frame:UnregisterAllEvents()
	if (p_t_c) then frame:RegisterEvent("PLAYER_TARGET_CHANGED") end
	if (p_f_c) then frame:RegisterEvent("PLAYER_FOCUS_CHANGED") end
	if (u_p) then frame:RegisterEvent("UNIT_PET") end
	if (u_t) then frame:RegisterEvent("UNIT_TARGET") end
	if (p_m_c) then frame:RegisterEvent("PARTY_MEMBERS_CHANGED") end
	if (r_r_u) then frame:RegisterEvent("RAID_ROSTER_UPDATE") end
end -- update_events()

local function add_tag_listener(tag, cb, handler)
	if (not tagSupported[tag]) then
		return false -- TODO: error()?
	end
	-- validate args and store the callback
	local first = false
	if (type(handler) == "table") then
		if (type(cb) ~= "string" and type(cb) ~= "function") then
			return false -- TODO: error()?
		elseif (tagMethods[tag] and tagMethods[tag][handler] and tagMethods[tag][handler][cb]) then
			return true
		end
		queueMethods[handler] = queueMethods[handler] or {}
		tagMethods[tag] = tagMethods[tag] or {}
		if (not tagMethods[tag][handler]) then
			first = not tagFunctions[tag]
			tagMethods[tag][handler] = {}
		end
		tagMethods[tag][handler][cb] = true
	else
		if (type(cb) ~= "function") then
			return false -- TODO: error()?
		elseif (tagFunctions[tag] and tagFunctions[tag][cb]) then
			return true
		end
		if (not tagFunctions[tag]) then
			first = not tagMethods[tag]
			tagFunctions[tag] = {}
		end
		tagFunctions[tag][cb] = true
	end
	-- if this is the tag's first listener, establish identities
	if (first) then
		tagTracked[tag] = true
		tagID[tag] = API_UnitGUID(tag)
		local parent = tagParent[tag]
		while (parent) do
			tagTracked[parent] = true
			tagID[parent] = tagID[parent] or API_UnitGUID(parent)
			parent = tagParent[parent]
		end
		update_events()
	end
	return true
end -- add_listener()

local update_tag_watch
update_tag_watch = function(tag)
	-- recurse to child tags first, since one of them might get cleaned up even if we still have a listener
	-- don't need player=>vehicle because we'll never call this on player anyway
	local pet_watched = (tagPet[tag] and update_tag_watch(tagPet[tag]))
	local target_watched = (tagTarget[tag] and update_tag_watch(tagTarget[tag]))
	-- but if nobody's listening, stop watching
	if (not (tagFunctions[tag] or tagMethods[tag] or pet_watched or target_watched)) then
		tagTracked[tag] = false
		tagID[tag] = nil
		return false
	end
	return true
end -- update_tag_watch()

local function remove_tag_listener(tag, cb, handler)
	-- validate args and erase the callback
	local last = false
	if (type(handler) == "table") then
		if (not (tagMethods[tag] and tagMethods[tag][handler] and tagMethods[tag][handler][cb])) then
			return true
		end
		tagMethods[tag][handler][cb] = nil
		if (not next(tagMethods[tag][handler])) then
			tagMethods[tag][handler] = nil
			if (not next(tagMethods[tag])) then
				last = not tagFunctions[tag]
				tagMethods[tag] = nil
			end
			-- check if the handler is still used anywhere else
			local used = false
			for _,handlers in pairs(tagMethods) do
				for _h,_ in pairs(handlers) do
					if (_h == handler) then
						used = true
						break
					end
				end
				if (used) then break end
			end
			if (not used) then
				queueMethods[handler] = nil
			end
		end
	else
		if (not (tagFunctions[tag] and tagFunctions[tag][cb])) then
			return true
		end
		tagFunctions[tag][cb] = nil
		if (not next(tagFunctions[tag])) then
			last = not tagMethods[tag]
			tagFunctions[tag] = nil
		end
	end
	-- if this was the tag's last listener, clean up
	if (last) then
		while (tagParent[tag]) do
			tag = tagParent[tag]
		end
		update_tag_watch(tag)
		update_events()
	end
	return true
end -- remove_listener()

local update_unit
update_unit = function(tag)
	if (not tagTracked[tag]) then return true end
	local newID = API_UnitGUID(tag)
	if (newID == tagID[tag]) then return newID end
	-- record the change
	tagID[tag] = newID
	deltaSize = deltaSize + 1
	delta[deltaSize] = tag
	-- cascade to child tags
	-- don't need player=>vehicle because we'll never call this on player anyway
	if (tagTarget[tag]) then update_unit(tagTarget[tag]) end
	if (tagPet[tag]) then update_unit(tagPet[tag]) end
	return newID
end -- update_unit()

local function compile_update_unit_tests()
	-- return two test functions: one where the update short circuits on no-change, and one where it doesn't
	local testFunc
	do
		local tagTracked = { target=1 }
		local API_UnitGUID = API_UnitGUID
		local tagID = {}
		testFunc = function(tag) -- thou shalt only call this with the argument "target"!
			if (not tagTracked[tag]) then return true end
			local newID = API_UnitGUID(tag)
			if (newID == tagID[tag]) then return newID end
			tagID[tag] = newID -- this is only done once, so in a loop of thousands should have negligable affect on the total time
		end
	end
	local updateFunc
	do
		local tagTracked = { target=1 }
		local API_UnitGUID = API_UnitGUID
		local tagID = { target=1 }
		local _tagID = { target=1 }
		local deltaSize = 1
		local _deltaSize = 0
		local delta = {}
		local tagTarget = {}
		local tagPet = {}
		updateFunc = function(tag)
			if (not tagTracked[tag]) then return true end
			local newID = API_UnitGUID(tag)
			if (newID == tagID[tag]) then return newID end
			_tagID[tag] = newID
			_deltaSize = _deltaSize + 1
			delta[deltaSize] = tag
			-- recursive calls are still logged, so we only need to measure if...end overhead here
			if (tagTarget[tag]) then error() end
			if (tagPet[tag]) then error() end
			return newID
		end
	end
	return testFunc,updateFunc
end -- compile_update_unit_tests()

local function dispatch_callbacks(event)
	-- localize upvalues
	local pairs,pcall,type,wipe = pairs,pcall,type,wipe
	local tagFunctions,tagMethods = tagFunctions,tagMethods
	local queueFunctions,queueMethods = queueFunctions,queueMethods
	local tagID,delta,deltaWrapper,dS = tagID,delta,deltaWrapper,deltaSize
	local dS = deltaSize -- rename, because we need to change the upvalue itself at the end
	local tag
	-- add hash component to delta table, and find applicable callbacks
	for i = 1,dS do
		tag = delta[i]
		delta[tag] = tagID[tag]
		if (tagFunctions[tag]) then
			for cb,_ in pairs(tagFunctions[tag]) do
				queueFunctions[cb] = true
			end
		end
		if (tagMethods[tag]) then
			for handler,cbs in pairs(tagMethods[tag]) do
				-- queueMethods[handler] should already exist!
				for cb,_ in pairs(cbs) do
					if (type(cb) == "string") then
						cb = handler[cb]
						if (type(cb) == "function") then
							queueMethods[handler][cb] = true
						else
							-- TODO: error()? print()? how to notify without stopping other callbacks?
						end
					else
						queueMethods[handler][cb] = true
					end
				end
			end
		end
	end
	delta[0] = dS
	-- run function callbacks
	for cb,_ in pairs(queueFunctions) do
		pcall(cb, deltaWrapper, event)
	end
	wipe(queueFunctions)
	-- run method callbacks
	for handler,cbs in pairs(queueMethods) do
		for cb,_ in pairs(cbs) do
			pcall(cb, handler, deltaWrapper, event)
		end
		wipe(cbs)
	end
	deltaSize = 0
end -- dispatch_callbacks()


--[[ EVENT HANDLER ]]--

frame:SetScript("OnEvent", function(self, event, arg1)
	-- update affected unit(s)
--print(event.."("..tostring(arg1)..")")
	if (event == "UNIT_TARGET") then
		update_unit(tagTarget[arg1] or false)
	elseif (event == "PLAYER_TARGET_CHANGED") then
		update_unit("target")
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		update_unit("focus")
	elseif (event == "RAID_ROSTER_UPDATE") then
		local uu = update_unit
		uu("raid1");  uu("raid2");  uu("raid3");  uu("raid4");  uu("raid5")
		uu("raid6");  uu("raid7");  uu("raid8");  uu("raid9");  uu("raid10")
		uu("raid11"); uu("raid12"); uu("raid13"); uu("raid14"); uu("raid15")
		uu("raid16"); uu("raid17"); uu("raid18"); uu("raid19"); uu("raid20")
		uu("raid21"); uu("raid22"); uu("raid23"); uu("raid24"); uu("raid25")
		uu("raid26"); uu("raid27"); uu("raid28"); uu("raid29"); uu("raid20")
		uu("raid31"); uu("raid32"); uu("raid33"); uu("raid34"); uu("raid35")
		uu("raid36"); uu("raid37"); uu("raid38"); uu("raid39"); uu("raid40")
	elseif (event == "PARTY_MEMBERS_CHANGED") then
		local uu = update_unit
		uu("party1"); uu("party2"); uu("party3"); uu("party4"); uu("party5")
	elseif (event == "UNIT_PET") then
		update_unit(tagPet[arg1] or false)
		if (arg1 == "player") then
			update_unit("vehicle")
		end
	end
	if (deltaSize > 0) then
		return dispatch_callbacks(event)
	end
end)


--[[ STATISTICS & ANALYSIS ]]--
-- undocumented, but needed by the disembedded utilities

local real_update_unit = update_unit
local function log_update_unit(tag)
	logTests = logTests + 1
	return real_update_unit(tag)
end -- log_update_unit()

local real_dispatch_callbacks = dispatch_callbacks
local function log_dispatch_callbacks(event)
	logUpdates = logUpdates + deltaSize
	return real_dispatch_callbacks(event)
end -- log_dispatch_callbacks()

local real_UnitGUID -- stored on the first logging call
local function log_UnitGUID(unitid)
	logQueries = logQueries + 1
	return real_UnitGUID(unitid)
end -- log_UnitGUID()

local real_GetUnitGUID -- stored on the first logging call
local function log_GetUnitGUID(self, unitid)
	logQueries = logQueries + 1
	return real_GetUnitGUID(self, unitid)
end -- log_GetUnitGUID()

function LibUnitID:__start_logging()
	logActive = true
	real_GetUnitGUID = real_GetUnitGUID or self.GetUnitGUID
	real_UnitGUID = real_UnitGUID or self.UnitGUID
	update_unit = log_update_unit
	dispatch_callbacks = log_dispatch_callbacks
	self.GetUnitGUID = log_GetUnitGUID
	self.UnitGUID = log_UnitGUID
end -- __start_logging()

function LibUnitID:__get_statistics()
	return logTests,logUpdates,logQueries
end -- __get_statistics()

function LibUnitID:__reset_statistics()
	logTests = 0
	logUpdates = 0
	logQueries = 0
end -- __reset_statistics()

function LibUnitID:__stop_logging()
	logActive = false
	update_unit = real_update_unit
	dispatch_callbacks = real_dispatch_callbacks
	self.GetUnitGUID = real_GetUnitGUID
	self.UnitGUID = real_UnitGUID
end -- __stop_logging()

function LibUnitID:__get_speed_tests()
	return compile_update_unit_tests()
end -- __get_speed_tests()


--[[ RUNTIME UPGRADE ]]--

-- undocumented, for internal use only!
function LibUnitID:__dump_for_upgrade()
	return {
		frame = frame,
		tagFunctions = tagFunctions,
		tagMethods = tagMethods,
		logActive = logActive,
		logTests = logTests,
		logUpdates = logUpdates,
		logQueries = logQueries
	}
end -- __dump_for_upgrade()

if (olddata.logActive) then
	LibUnitID:__start_logging()
end

if (olddata.tagListeners) then -- pre v.9
	olddata.tagFunctions = {}
	olddata.tagMethods = {}
	for tag,cbs in pairs(olddata.tagListeners) do
		for cb,handler in pairs(cbs) do
			if (type(handler) == "table") then
				olddata.tagMethods[tag] = olddata.tagMethods[tag] or {}
				olddata.tagMethods[tag][handler] = olddata.tagMethods[tag][handler] or {}
				olddata.tagMethods[tag][handler][cb] = true
			else
				olddata.tagFunctions[tag] = olddata.tagFunctions[tag] or {}
				olddata.tagFunctions[tag][cb] = true
			end
		end
	end
end

if (olddata.tagFunctions or olddata.tagMethods) then
	local ok
	local n,e = 0,0
	if (olddata.tagFunctions) then
		for tag,cbs in pairs(olddata.tagFunctions) do
			for cb,_ in pairs(cbs) do
				ok = add_tag_listener(tag, cb, handler)
				if (ok) then
					n = n + 1
				else
					e = e + 1
					print("LibUnitID ERROR: Failed to re-register function callback for unit '"..tag.."' during library upgrade!")
				end
			end
			wipe(cbs)
		end
		wipe(olddata.tagFunctions)
		olddata.tagFunctions = nil
	end
	if (olddata.tagMethods) then
		for tag,handlers in pairs(olddata.tagMethods) do
			for handler,cbs in pairs(handlers) do
				for cb,_ in pairs(cbs) do
					ok = add_tag_listener(tag, cb, handler)
					if (ok) then
						n = n + 1
					else
						e = e + 1
						print("LibUnitID ERROR: Failed to re-register method callback for unit '"..tag.."' during library upgrade!")
					end
				end
				wipe(cbs)
			end
			wipe(handlers)
		end
		wipe(olddata.tagMethods)
		olddata.tagMethods = nil
	end
--[===[@debug@
	print("LibUnitID: Upgraded from revision "..olddata.MINOR.."; "..n.." callbacks re-registered, "..e.." errors.")
elseif (olddata.MINOR) then
	print("LibUnitID: Upgraded from revision "..olddata.MINOR.."; no callbacks re-registered.")
--@end-debug@]===]
end


--[[ PUBLIC API METHODS ]]--

--=== compatibility ===--

--- Gets a list of the UnitIDs that the library is able to track and cache.
-- The returned table contains the supported UnitIDs as both keys and values,
-- such as '''{target="target",pet="pet",...}'''.  These are the UnitIDs that
-- may be registered for callbacks using ''':RegisterUnit()''', and then
-- queried using ''':GetUnitGUID()''' or '''.UnitGUID()'''.
-- @return A hash table containing the supported UnitIDs.
function LibUnitID:GetSupportedUnitIDs()
	-- build a fresh table, in case the caller wants to muck it up
	local t = {}
	for tag,_ in pairs(tagSupported) do
		t[tag] = tag
	end
	return t
end -- GetSupportedUnitIDs()

--=== register unit ===--

--- Registers a callback for a single UnitID.
-- Whenever the identity of the specified UnitID changes, the callback
-- function will be called.  If a handler is provided, it must be a table and
-- will be given as the first argument to the callback.  The next arguments to
-- the callback will be the name of the game event that triggered the update,
-- followed by a '''delta''' table detailing the UnitIDs that changed.
--
-- If a handler is provided, then '''callback''' may be specified as a string,
-- which will be resolved on the handler each time the callback is to be
-- called.  This is slightly slower but more flexible, since it allows you to
-- redefine the callback function without having to unregister and re-register
-- the function pointer.
--
-- Multiple calls with the same '''unitid''', '''callback''' and '''handler'''
-- have no effect.  A single '''callback''' may be registered with multiple
-- '''handler'''s, and a single '''handler''' may be registered with multiple
-- '''callback'''s, and each combination will be called.
--
-- If the same '''callback'''+'''handler''' pair is registered for multiple
-- UnitIDs, and if a single game event causes more than one of those UnitIDs
-- to change identity, the callback will still only be called once.  To
-- retrieve information about which UnitIDs actually changed within your
-- callback, check the provided '''delta''' table.
-- @see delta
-- @param unitid The UnitID to monitor.
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return True if the callback is successfully registered, false otherwise.
function LibUnitID:RegisterUnit(unitid, callback, handler)
	return add_tag_listener(unitid, callback, handler)
end -- RegisterUnit()

--=== register party ===--

--- Registers a callback for all party-member UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs "party1"
-- through "party5".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterParty()

--- Registers a callback for all party-member-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "party1target" through "party5target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyTargets()

--- Registers a callback for all party-member-pet UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "partypet1" through "partypet5".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPets()

--- Registers a callback for all party-member-pet-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "partypet1target" through "partypet5target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPetTarges()

--- Registers a callback for all party related UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "party#", "party#target", "partypet#" and "partypet#target" for # 1..5.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (20 possible).
function LibUnitID:RegisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllParty()

--=== register raid ===--

--- Registers a callback for all raid-member UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs "raid1"
-- through "raid40".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaid()

--- Registers a callback for all raid-member-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raid1target" through "raid40target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidTargets()

--- Registers a callback for all raid-member-pet UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raidpet1" through "raidpet40".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPets()

--- Registers a callback for all raid-member-pet-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raidpet1target" through "raidpet40target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPetTargets()

--- Registers a callback for all raid related UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raid##", "raid##target", "raidpet##" and "raidpet##target" for ## 1..40.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (160 possible).
function LibUnitID:RegisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllRaid()

--=== register all ===--

--- Registers a callback for all supported UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on all UnitIDs that
-- are supported by LibUnitID.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered.
function LibUnitID:RegisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAll()

--=== unregister unit ===--

--- Unregisters a callback from a single UnitID.
-- The callback function or method will no longer be called for changes in the
-- specified UnitID.  Callbacks are stored separately and must be unregistered
-- separately for each '''callback'''+'''handler''' pair, including whether
-- '''callback''' is a function reference or a string method name.
-- @param unitid The UnitID to stop monitoring.
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return True if the callback is successfully unregistered, false otherwise (including if the callback was not registered for the unitid).
function LibUnitID:UnregisterUnit(unitid, callback, handler)
	return remove_tag_listener(unitid, callback, handler)
end -- UnregisterUnit()

--=== unregister party ===--

--- Unregisters a callback from all party-member UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party1" through "party5".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterParty()

--- Unregisters a callback from all party-member-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party1target" through "party5target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyTargets()

--- Unregisters a callback from all party-member-pet UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "partypet1" through "partypet5".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPets()

--- Unregisters a callback from all party-member-pet-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "partypet1target" through "partypet5target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPetTargets()

--- Unregisters a callback from all party related UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party#", "party#target", "partypet#" and "partypet#target" for # 1..5.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (20 possible).
function LibUnitID:UnregisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllParty()

--=== unregister raid ===--

--- Unregisters a callback from all raid-member UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid1" through "raid40".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaid()

--- Unregisters a callback from all raid-member-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid1target" through "raid40target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidTargets()

--- Unregisters a callback from all raid-member-pet UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raidpet1" through "raidpet40".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPets()

--- Unregisters a callback from all raid-member-pet-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raidpet1target" through "raidpet40target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPetTargets()

--- Unregisters a callback from all raid related UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid#", "raid#target", "raidpet#" and "raidpet#target" for # 1..40.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (160 possible).
function LibUnitID:UnregisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllRaid()

--=== unregister all ===--

--- Unregisters a callback from all supported UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on all UnitIDs that
-- are supported by LibUnitID.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered.
function LibUnitID:UnregisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAll()

--=== guid queries ===--

--- Returns the UnitID's GUID from the cache.
-- This is the method-call version of this functionality; it must be called as
-- a method on a LibUnitID library reference, as in
-- '''myLibRef:GetUnitGUID("pet")'''.
--
-- Note that the cache is only updated when the library receives event
-- notifications from the game engine, which may happen in any order.  If you
-- have your own event handlers for events that affect UnitIDs, and if you try
-- to query the cache during those event handlers, you may receive out-of-date
-- information if the library has not yet received the event and updated
-- itself.
--
-- Also note that the library will not track UnitIDs for which nobody has
-- registered a callback, and will also not take the time to verify that when
-- queried; if you request the GUID of a UnitID that you did not register a
-- callback for, you will get '''nil''' unless someone else happened to
-- register for that UnitID.
--
-- For best results, only query the cache from a callback that you registered
-- with LibUnitID, or during events that do not affect UnitIDs.
-- @param unitid The UnitID to fetch the GUID for.
-- @return The GUID of the UnitID, or nil if it doesn't exist or isn't being tracked by the library.
function LibUnitID:GetUnitGUID(unitid)
	return tagID[unitid]
end -- GetUnitGUID()

--- Returns the UnitID's GUID from the cache.
-- This is the function-call version of this functionality; it must be called
-- as a bare function, as in '''myLibRef.UnitGUID("pet")'''.  This is done so
-- that you can make a local copy of the function reference (as in
-- '''local UnitGUID = myLibRef.UnitGUID''') and then use it in place of the
-- default API UnitGUID() function.  In all other ways this is equivalent to
-- ''':GetUnitGUID()''', and all the same caveats apply.
-- @see LibUnitID:GetUnitGUID
-- @param unitid The UnitID to fetch the GUID for.
-- @return The GUID of the UnitID, or nil if it doesn't exist or isn't being tracked by the library.
function LibUnitID.UnitGUID(unitid) -- NOT A METHOD!  that's a dot, not a colon, Jim!
	return tagID[unitid]
end -- UnitGUID()

