local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 9
local LibUnitID,libminor = LibStub(MAJOR,MINOR)

-- set a default chat command hook in case we bail out
SlashCmdList["LIBUNITID"] = function()
	print("LibUnitID: Utilities rev. "..MINOR.." are not compatible with rev. "..libminor..".  Please update your standalone !LibUnitID to use the utilities.")
end
SLASH_LIBUNITID1 = "/libunitid"
SLASH_LIBUNITID2 = "/lui"

-- bail out if a newer copy was already loaded somehow
if (libminor ~= MINOR) then
	SlashCmdList["LIBUNITID"]()
	return
end

-- check which addons already loaded, and thus probably already localized UnitGUID
local preloaded = {}
do
	local n = GetNumAddOns()
	if (n > 0) then
		for i=1,n do
			--local name,title,notes,enabled,loadable,reason,security = GetAddOnInfo(i)
			preloaded[n] = (GetAddOnInfo(i))
		end
	end
end

-- initialize vars
local timingRuns = 0
local timeAPI = 0
local timeCacheTest = 0
local timeCacheUpdate = 0
local timeCacheHit = 0
local logSince = GetTime()
local logCachableAPI = 0
local logUncachableAPI = 0

-- get the supported UnitIDs and speed test functions from the library
local tagSupported = LibUnitID:GetSupportedUnitIDs()
local libCacheTest,libCacheUpdate = LibUnitID:__get_speed_tests()

-- store initial function pointers
local API_UnitGUID = UnitGUID
local lib_GetUnitGUID = LibUnitID.GetUnitGUID
local lib_UnitGUID = LibUnitID.UnitGUID

-- set the traps (the relay mechanism is for rascally addons who localize function pointers)
local relay_API_UnitGUID = API_UnitGUID
UnitGUID = function(...) return relay_API_UnitGUID(...) end
_G["UnitGUID"] = function(...) return relay_API_UnitGUID(...) end

local relay_lib_UnitGUID = function(...) return LibUnitID.UnitGUID(...) end
LibUnitID.UnitGUID = relay_lib_UnitGUID

-- define the API wrapper
local log_API_UnitGUID = function(unitid, ...)
	if (tagSupported[unitid]) then
		logCachableAPI = logCachableAPI + 1
	else
		logUncachableAPI = logUncachableAPI + 1
	end
	return API_UnitGUID(unitid, ...)
end

-- monitor/eventhandler frame
local frame = CreateFrame("Frame", nil, UIParent)

local monitor_handler = {}

monitor_handler.monitor_callback = function(self, delta, event)
	if (self ~= monitor_handler) then
		print("LibUnitID: Bad method callback for " .. event)
	end
	if (LibUnitID_DB.verbose) then
		local n = delta[0]
		local s = ""
		for i=1,n do
			s = s .. ", " .. delta[i] .. "(" .. tostring(delta[delta[i]]) .. ")"
		end
		if (s == "") then
			print("LibUnitID: " .. event .. " => EMPTY UPDATE")
		else
			print("LibUnitID: " .. event .. " => " .. strsub(s,3))
		end
	end
end -- monitor_callback()

local function monitor_OnShow()
	-- we should still only see one call per event
	if (not LibUnitID:RegisterAll(monitor_handler.monitor_callback, monitor_handler)) then
		print("LibUnitID: Failed to register monitor handler function.")
	end
	if (not LibUnitID:RegisterAll("monitor_callback", monitor_handler)) then
		print("LibUnitID: Failed to register monitor handler method.")
	end
end -- monitor_OnShow()

local function monitor_OnHide()
	if (not LibUnitID:UnregisterAll(monitor_handler.monitor_callback, monitor_handler)) then
		print("LibUnitID: Failed to unregister monitor handler function.")
	end
	if (not LibUnitID:UnregisterAll("monitor_callback", monitor_handler)) then
		print("LibUnitID: Failed to unregister monitor handler method.")
	end
end -- monitor_OnHide()

-- when savedvars load, initialize logging and monitoring
frame:SetScript("OnEvent", function(self, event, arg1)
	if (arg1 == "!LibUnitID") then
		if (not LibUnitID_DB) then
			LibUnitID_DB = {}
		end
		if (LibUnitID_DB.log) then
			relay_API_UnitGUID = log_API_UnitGUID
			relay_lib_GetUnitGUID = log_lib_GetUnitGUID
			relay_lib_UnitGUID = log_lib_UnitGUID
		end
		frame:UnregisterEvent("ADDON_LOADED")
		frame:SetScript("OnEvent", nil)
		frame:Hide()
		frame:SetScript("OnShow", monitor_OnShow)
		frame:SetScript("OnHide", monitor_OnHide)
		if (LibUnitID_DB.monitor) then
			frame:Show()
		end
	end
end)
frame:RegisterEvent("ADDON_LOADED")


--[[ TIMING TEST ]]--

local function generate_timing_data()
	local GetTime = GetTime
	local newRuns = timingRuns + 1
	local n,t,func
	-- sanity check
	if (not API_UnitGUID("target")) then return false end
	-- timing loop overhead
	n = 128
	repeat
		n = n * 2
		t = GetTime()
		for i=1,n do end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	local timeLoop = (t * 1e9) / n -- nanoseconds per loop; not refined: we always want a current snapshot
	-- API call
	n = 128
	repeat
		local func = API_UnitGUID
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeAPI = ((timeAPI * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache test
	n = 128
	repeat
		local func = libCacheTest
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheTest = ((timeCacheTest * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache update
	n = 128
	repeat
		local func = libCacheUpdate
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheUpdate = ((timeCacheUpdate * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache hit
	n = 128
	repeat
		local func = lib_UnitGUID
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheHit = ((timeCacheHit * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- done
	timingRuns = newRuns
	return true
end -- generate_timing_data()


--[[ COMMAND PARSER ]]--

local function s(f) return (f<0.001 and format("%.3f",f*1e6).."us") or (f<1 and format("%.3f",f*1e3).."ms") or format("%.3f",f).."s" end
local function ms(f) return (f<1 and format("%.3f",f*1e3).."us") or (f<1e3 and format("%.3f",f).."ms") or format("%.3f",f/1e3).."s" end
local function us(f) return (f<1e3 and format("%.3f",f).."us") or (f<1e6 and format("%.3f",f/1e3).."ms") or format("%.3f",f/1e6).."s" end
local function ns(f) return (f<1e3 and format("%.3f",f).."ns") or (f<1e6 and format("%.3f",f/1e3).."us") or format("%.3f",f/1e6).."ms" end
local function pct(f) return format("%.3f",f).."%" end

SlashCmdList["LIBUNITID"] = function(args)
	-- make sure the library didn't upgrade
	local LibUnitID,libminor = LibStub(MAJOR,MINOR)
	if (libminor ~= MINOR) then
		print("LibUnitID: Utilities rev. "..MINOR.." are not compatible with rev. "..libminor..".  Please update your standalone !LibUnitID to use the utilities.")
		return
	end
	-- process command
	if (strsub(args,1,3) == "log") then
		local log = strsub(args,5)
		if (log == "") then
			print("LibUnitID: Logging is currently "..((LibUnitID_DB.log and "on") or "off")..".")
		elseif (log == "on") then
			if (LibUnitID_DB.log) then
				print("LibUnitID: Logging is already on.")
			else
				print("LibUnitID: GUID queries and library cache upkeep are now logged.")
				if (#preloaded > 0) then print("LibUnitID: "..#preloaded.." addons loaded before !LibUnitID, and their GUID queries might bypass the logging hooks; check '/libunitid list'.") end
				relay_API_UnitGUID = log_API_UnitGUID
				relay_lib_GetUnitGUID = log_lib_GetUnitGUID
				relay_lib_UnitGUID = log_lib_UnitGUID
				LibUnitID:__start_logging()
				LibUnitID_DB.log = true
			end
		elseif (log == "off") then
			if (LibUnitID_DB.log) then
				print("LibUnitID: GUID queries and library cache upkeep are no longer logged.")
				relay_API_UnitGUID = API_UnitGUID
				relay_lib_GetUnitGUID = lib_GetUnitGUID
				relay_lib_UnitGUID = lib_UnitGUID
				LibUnitID:__stop_logging()
				LibUnitID_DB.log = nil
			else
				print("LibUnitID: Logging is already off.")
			end
		end
	elseif (args == "time") then
		print("LibUnitID: Generating timing data...")
		if (generate_timing_data()) then
			print("...done.  Results aggregated over "..timingRuns.." run"..((timingRuns>1 and "s") or "")..".")
			print("Nanoseconds per operation:")
			print("  "..ns(timeAPI).." per API UnitGUID()")
			print("  "..ns(timeCacheTest).." per Cache Change Test")
			print("  "..ns(timeCacheUpdate).." per Cache Identity Update")
			print("  "..ns(timeCacheHit).." per Cached GUID Query")
		else
			print("...aborted: You must have a target to generate timing data; target something and try again.")
		end
    elseif (args == "report") then
		local interval = GetTime() - logSince -- seconds
		print("LibUnitID: Operations in the last "..s(interval)..":")
		local logCacheTests,logCacheUpdates,logCacheQueries = LibUnitID:__get_statistics()
		print("  "..logCachableAPI.." Cachable API UnitGUID()s")
		print("  "..logUncachableAPI.." Uncachable API UnitGUID()s")
		print("  "..logCacheTests.." Cache Change Tests")
		print("  "..logCacheUpdates.." Cache Identity Updates")
		print("  "..logCacheQueries.." Cached GUID Queries")
		if (timingRuns > 0) then
			print("Nanoseconds per operation:")
			print("  "..ns(timeAPI).." per API UnitGUID()")
			print("  "..ns(timeCacheTest).." per Cache Change Test")
			print("  "..ns(timeCacheUpdate).." per Cache Identity Update")
			print("  "..ns(timeCacheHit).." per Cached GUID Query")
			print("Analysis:")
			local cost = ((logCacheTests * timeCacheTest) / 1e3) + ((logCacheUpdates * timeCacheUpdate) / 1e3)  -- milliseconds
			print("  "..ms(cost).." spent maintaining the library's cache")
			local benefit = ((logCacheQueries * timeAPI) / 1e3) - ((logCacheQueries * timeCacheHit) / 1e3) -- milliseconds
			print("  "..ms(benefit).." saved by querying the cache instead of the API")
			local net = benefit - cost -- milliseconds
			local efficiency = ((net / 1e3) * 100.0) / interval
			if (net < 0) then
				print("Overall loss is "..pct(efficiency).." ("..ms(-net).." extra spent over the last "..s(interval)..")")
			else
				print("Overall gain is "..pct(efficiency).." ("..ms(net).." saved over the last "..s(interval)..")")
			end
			local bestBenefit = ((logCachableAPI * timeAPI) / 1e3) - ((logCachableAPI * timeCacheHit) / 1e3) -- milliseconds
			print("  "..ms(bestBenefit).." more could have been saved if cachable UnitGUID() calls had used the cache")
			local bestNet = net + bestBenefit -- milliseconds
			local bestEfficiency = ((bestNet / 1e3) * 100.0) / interval
			if (bestNet < 0) then
				print("Best-case loss is "..pct(bestEfficiency).." ("..ms(-bestNet).." extra still spent over the last "..s(interval)..")")
			else
				print("Best-case gain is "..pct(bestEfficiency).." ("..ms(bestNet).." could have been saved over the last "..s(interval)..")")
			end
		else
			print("Efficiency analysis not available without timing data; run '/libunitid time'")
		end
	elseif (args == "reset") then
		timingRuns = 0
		timeAPI = 0
		timeCacheTest = 0
		timeCacheUpdate = 0
		timeCacheHit = 0
		logSince = GetTime()
		logCachableAPI = 0
		logUncachableAPI = 0
		LibUnitID:__reset_statistics()
		print("LibUnitID: Statistics reset.")
	elseif (args == "list") then
		local n = #preloaded
		if (n == 0) then
			print("LibUnitID: No addons loaded before !LibUnitID; all calls should be trapped.")
		else
			print("LibUnitID: "..n.." addons loaded before !LibUnitID, and their GUID queries (both native-API and library-cached) might bypass the logging hooks:")
			local s = ""
			for n,name in ipairs(preloaded) do
				s = s..", "..name
			end
			print("  "..strsub(s,3))
		end
	elseif (strsub(args,1,7) == "monitor") then
		local monitor = strsub(args,9)
		if (monitor == "") then
			print("LibUnitID: Monitoring is currently "..((LibUnitID_DB.monitor and "on") or "off")..".")
		elseif (monitor == "on") then
			if (LibUnitID_DB.monitor) then
				print("LibUnitID: Monitoring is already on.")
			else
				print("LibUnitID: All supported unitids are now monitored.")
				frame:Show()
				LibUnitID_DB.monitor = true
			end
		elseif (monitor == "off") then
			if (LibUnitID_DB.monitor) then
				print("LibUnitID: Monitoring disabled.")
				frame:Hide()
				LibUnitID_DB.monitor = nil
			else
				print("LibUnitID: Monitoring is already off.")
			end
		end
	elseif (strsub(args,1,7) == "verbose") then
		local verbose = strsub(args,9)
		if (verbose == "") then
			print("LibUnitID: Verbose monitoring mode is currently "..((LibUnitID_DB.verbose and "on") or "off")..".")
		elseif (verbose == "on") then
			if (LibUnitID_DB.verbose) then
				print("LibUnitID: Monitoring is already verbose.")
			else
				print("LibUnitID: Monitoring is now verbose.")
				LibUnitID_DB.verbose = true
			end
		elseif (verbose == "off") then
			if (LibUnitID_DB.verbose) then
				print("LibUnitID: Monitoring is no longer verbose.")
				LibUnitID_DB.verbose = nil
			else
				print("LibUnitID: Monitoring is already not verbose.")
			end
		end
	elseif (args == "help") then
		print("LibUnitID commands via '/libunitid' or '/lui':")
		print("  /libunitid list  --  list addons that loaded before !LibUnitID, and might bypass logging")
		print("  /libunitid log [on | off]  --  log calls to API's and library's UnitGUID()")
		print("  /libunitid monitor [on | off]  --  monitor all supported UnitIDs (forces maximum cache upkeep work)")
		print("  /libunitid verbose [on | off]  --  when monitoring, also print all identity changes (spammy!)")
		print("  /libunitid time  --  generate or refine timing data for performance analysis")
		print("  /libunitid report  --  show logging results and performance analysis, if available")
		print("  /libunitid reset  --  reset call log counters and timing data")
		print("  /libunitid help  --  this")
    else
		print(MAJOR.." rev. "..MINOR.." standalone utilities")
		print("  Logging is currently "..((LibUnitID_DB.log and "ON") or "OFF")..".")
		print("  Monitoring is currently "..((LibUnitID_DB.monitor and "ON") or "OFF")..".")
		print("  Verbose monitoring is currently "..((LibUnitID_DB.verbose and "ON") or "OFF")..".")
		print("'/libunitid help' for usage.")
    end
end

